@@ -279,14 +279,23 @@ show_volume(LDM *const ldm, const gint argc, gchar ** const argv,
         LDMVolume * const vol = g_array_index(volumes, LDMVolume *, i);
 
         gchar *name = ldm_volume_get_name(vol);
-        LDMVolumeType type = ldm_volume_get_voltype(vol);
-        guint64 size = ldm_volume_get_size(vol);
-        guint64 chunk_size = ldm_volume_get_chunk_size(vol);
-        gchar *hint = ldm_volume_get_hint(vol);
-
         if (g_strcmp0(name, argv[1]) == 0) {
             found = TRUE;
 
+            gchar *guid = ldm_volume_get_guid(vol);
+            LDMVolumeType type = ldm_volume_get_voltype(vol);
+            guint64 size = ldm_volume_get_size(vol);
+            guint64 chunk_size = ldm_volume_get_chunk_size(vol);
+            gchar *hint = ldm_volume_get_hint(vol);
+
+            GError *err = NULL;
+            gchar *device = ldm_volume_dm_get_device(vol, &err);
+            if (err) {
+                g_warning("Unable to get device for volume %s with GUID %s: %s",
+                          name, guid, err->message);
+                g_error_free(err);
+            }
+
             json_builder_begin_object(jb);
 
             GEnumValue * const type_v =
@@ -294,6 +303,8 @@ show_volume(LDM *const ldm, const gint argc, gchar ** const argv,
 
             json_builder_set_member_name(jb, "name");
             json_builder_add_string_value(jb, name);
+            json_builder_set_member_name(jb, "guid");
+            json_builder_add_string_value(jb, guid);
             json_builder_set_member_name(jb, "type");
             json_builder_add_string_value(jb, type_v->value_nick);
             json_builder_set_member_name(jb, "size");
@@ -304,6 +315,10 @@ show_volume(LDM *const ldm, const gint argc, gchar ** const argv,
                 json_builder_set_member_name(jb, "hint");
                 json_builder_add_string_value(jb, hint);
             }
+            if (device != NULL) {
+                json_builder_set_member_name(jb, "device");
+                json_builder_add_string_value(jb, device);
+            }
 
             json_builder_set_member_name(jb, "partitions");
             json_builder_begin_array(jb);
@@ -320,10 +335,13 @@ show_volume(LDM *const ldm, const gint argc, gchar ** const argv,
             json_builder_end_array(jb);
 
             json_builder_end_object(jb);
+
+            g_free(guid);
+            g_free(hint);
+            g_free(device);
         }
 
         g_free(name);
-        g_free(hint);
 
         if (found) break;
     }
@@ -343,21 +361,29 @@ show_partition(LDM *const ldm, const gint argc, gchar ** const argv,
 
     GArray * const parts = ldm_disk_group_get_partitions(dg);
     g_object_unref(dg);
+    gboolean found = FALSE;
     for (guint i = 0; i < parts->len; i++) {
         LDMPartition * const part = g_array_index(parts, LDMPartition *, i);
 
         gchar *name = ldm_partition_get_name(part);
-        guint64 start = ldm_partition_get_start(part);
-        guint64 size = ldm_partition_get_size(part);
-
-        gboolean found = FALSE;
         if (g_strcmp0(name, argv[1]) == 0) {
             found = TRUE;
 
+            guint64 start = ldm_partition_get_start(part);
+            guint64 size = ldm_partition_get_size(part);
+
             LDMDisk * const disk = ldm_partition_get_disk(part);
             gchar *diskname = ldm_disk_get_name(disk);
             g_object_unref(disk);
 
+            GError *err = NULL;
+            gchar *device = ldm_partition_dm_get_device(part, &err);
+            if (err) {
+                g_warning("Unable to get device for partition %s on disk %s: %s",
+                          name, diskname, err->message);
+                g_error_free(err);
+            }
+
             json_builder_begin_object(jb);
 
             json_builder_set_member_name(jb, "name");
@@ -368,10 +394,15 @@ show_partition(LDM *const ldm, const gint argc, gchar ** const argv,
             json_builder_add_int_value(jb, size);
             json_builder_set_member_name(jb, "disk");
             json_builder_add_string_value(jb, diskname);
+            if (device != NULL) {
+                json_builder_set_member_name(jb, "device");
+                json_builder_add_string_value(jb, device);
+            }
 
             json_builder_end_object(jb);
 
             g_free(diskname);
+            g_free(device);
         }
 
         g_free(name);
@@ -380,7 +411,7 @@ show_partition(LDM *const ldm, const gint argc, gchar ** const argv,
     }
     g_array_unref(parts);
 
-    return TRUE;
+    return found;
 }
 
 gboolean
@@ -399,16 +430,16 @@ show_disk(LDM *const ldm, const gint argc, gchar ** const argv,
         LDMDisk * const disk = g_array_index(disks, LDMDisk *, i);
 
         gchar *name = ldm_disk_get_name(disk);
-        gchar *guid = ldm_disk_get_guid(disk);
-        gchar *device = ldm_disk_get_device(disk);
-        guint64 data_start = ldm_disk_get_data_start(disk);
-        guint64 data_size = ldm_disk_get_data_size(disk);
-        guint64 metadata_start = ldm_disk_get_metadata_start(disk);
-        guint64 metadata_size = ldm_disk_get_metadata_size(disk);
-
         if (g_strcmp0(name, argv[1]) == 0) {
             found = TRUE;
 
+            gchar *guid = ldm_disk_get_guid(disk);
+            gchar *device = ldm_disk_get_device(disk);
+            guint64 data_start = ldm_disk_get_data_start(disk);
+            guint64 data_size = ldm_disk_get_data_size(disk);
+            guint64 metadata_start = ldm_disk_get_metadata_start(disk);
+            guint64 metadata_size = ldm_disk_get_metadata_size(disk);
+
             json_builder_begin_object(jb);
 
             json_builder_set_member_name(jb, "name");
@@ -431,17 +462,18 @@ show_disk(LDM *const ldm, const gint argc, gchar ** const argv,
             }
 
             json_builder_end_object(jb);
+
+            g_free(guid);
+            g_free(device);
         }
 
         g_free(name);
-        g_free(guid);
-        g_free(device);
 
         if (found) break;
     }
     g_array_unref(disks);
 
-    return TRUE;
+    return found;
 }
 
 gboolean
