@@ -61,7 +61,7 @@ int main(int argc, const char *argv[])
             g_object_get(dg, "guid", &guid, "name", &name, NULL);
 
             printf("Disk Group: %s\n", name);
-            printf("  GUID: %s\n", guid);
+            printf("  GUID:   %s\n", guid);
 
             g_free(guid);
             g_free(name);
@@ -73,6 +73,7 @@ int main(int argc, const char *argv[])
 
             {
                 gchar *name;
+                gchar *guid;
                 LDMVolumeType type;
                 guint64 size;
                 guint32 part_type;
@@ -80,23 +81,31 @@ int main(int argc, const char *argv[])
                 guint64 chunk_size;
 
                 g_object_get(vol, "name", &name, "type", &type,
-                                  "size", &size, "part-type", &part_type,
-                                  "hint", &hint, "chunk-size", &chunk_size,
-                                  NULL);
+                                  "guid", &guid, "size", &size,
+                                  "part-type", &part_type, "hint", &hint,
+                                  "chunk-size", &chunk_size, NULL);
 
                 GEnumValue * const type_v =
                     g_enum_get_value(g_type_class_peek(LDM_TYPE_VOLUME_TYPE),
                                      type);
 
+                GError *err = NULL;
+                gchar *device = ldm_volume_dm_get_device(vol, &err);
+
                 printf("  Volume: %s\n", name);
+                printf("    GUID:       %s\n", guid);
                 printf("    Type:       %s\n", type_v->value_nick);
                 printf("    Size:       %lu\n", size);
                 printf("    Part Type:  %hhu\n", part_type);
                 printf("    Hint:       %s\n", hint);
                 printf("    Chunk Size: %lu\n", chunk_size);
+                printf("    Device:     %s\n", device);
 
                 g_free(name);
+                g_free(guid);
                 g_free(hint);
+                g_free(device);
+                if (err) g_error_free(err);
             }
 
             GArray *parts = ldm_volume_get_partitions(vol);
@@ -112,11 +121,17 @@ int main(int argc, const char *argv[])
                     g_object_get(part, "name", &name, "start", &start,
                                        "size", &size, NULL);
 
+                    GError *err = NULL;
+                    gchar *device = ldm_partition_dm_get_device(part, &err);
+
                     printf("    Partition: %s\n", name);
-                    printf("      Start: %lu\n", start);
-                    printf("      Size:  %lu\n", size);
+                    printf("        Start:  %lu\n", start);
+                    printf("        Size:   %lu\n", size);
+                    printf("        Device: %s\n", device);
 
                     g_free(name);
+                    g_free(device);
+                    if (err) g_error_free(err);
                 }
 
                 LDMDisk * const disk = ldm_partition_get_disk(part);
@@ -138,13 +153,13 @@ int main(int argc, const char *argv[])
                                        "metadata-size", &metadata_size,
                                        NULL);
 
-                    printf("      Disk: %s\n", name);
-                    printf("        GUID:   %s\n", guid);
-                    printf("        Device: %s\n", device);
-                    printf("        Data Start: %lu\n", data_start);
-                    printf("        Data Size: %lu\n", data_size);
-                    printf("        Metadata Start: %lu\n", metadata_start);
-                    printf("        Metadata Size: %lu\n", metadata_size);
+                    printf("        Disk: %s\n", name);
+                    printf("          GUID:           %s\n", guid);
+                    printf("          Device:         %s\n", device);
+                    printf("          Data Start:     %lu\n", data_start);
+                    printf("          Data Size:      %lu\n", data_size);
+                    printf("          Metadata Start: %lu\n", metadata_start);
+                    printf("          Metadata Size:  %lu\n", metadata_size);
 
                     g_free(name);
                     g_free(guid);
